#include "MOTestFunctions.h"

//------------------------------------------------------------------------------
// Class MOTestFunctions
//-------------------------------------------------
// Protected Members

/* (Slightly adjusted) DTLZ1 implementation from PISA (www.tik.ee.ethz.ch/pisa/),
 * dtlz_c_source by Marco Laumanns, laumanns@tik.ee.ethz.ch
 *
 * Domain: [0,1]^n, n >=7, (see Deb, 2002: Scalable MO Test Problems)
 * reference point (0.7,0.7,0.7) (see Deb, 2005: Evaluating the e-Domination Based MOEA)
 */
void MOTestFunctions::DTLZ1(Individual* I, double* xreal, int n, double* inputNoise)
{
#ifdef DEBUG
  cout << "MOTestFunctions::DTLZ1()" << endl;
#endif

  if (n < 7)  // Number of decision variables, recommended value for k >= 5 (see Deb, 2002: Scalable MO Test Problems), 
  {           // hence n >= 7
    cerr << "DTLZ1: number of decision variables, n = " << n <<  ", but should be n >= 7 (then  k >= 5)! Exiting." << endl;
    exit(1);
  }
      
  int M = 3;         // Problem dimension, i.e., M-objective
  int k = n - M + 1; // Length of xM input vector of g(xM), i.e., xM = (xM_0,...,xM_k-1)
 
  F.assign(M, 0);
   
  double g = 0;
  for (int i = M; i <= n; i++)
    g += pow(((I != NULL) ? I->R[i-1] : xreal[i-1]) + ((inputNoise != NULL) ? inputNoise[i-1] : 0) 
           - 0.5, 2) - cos(20 * PI * (((I != NULL) ? I->R[i] : xreal[i]) - 0.5));
  g = 100 * (k + g);
    
  for (int i = 1; i <= M; i++)
  {
    double f = 0.5 * (1 + g);
    for (int j = M - i; j >= 1; j--)
      f *= ((I != NULL) ? I->R[j-1] : xreal[j-1]) + ((inputNoise != NULL) ? inputNoise[j-1] : 0);
      
    if (i > 1)
    {
      int index = (M - i + 1) - 1;
      f *= 1 - ((I != NULL) ? I->R[index] : xreal[index]) + ((inputNoise != NULL) ? inputNoise[index] : 0);
    }

    F[i-1] = f;
  }
}

/* (Slightly adjusted) DTLZ2 implementation from PISA (www.tik.ee.ethz.ch/pisa/),
 * dtlz_c_source by Marco Laumanns, laumanns@tik.ee.ethz.ch
 *
 * Domain: [0,1]^n, n >= 12, (see Deb, 2002: Scalable MO Test Problems)
 * reference point (1.1,1.1,1.1) (see Naujoks, 2005: Multi-objective Optimisation Using S-Metric) 
 */
void MOTestFunctions::DTLZ2(Individual* I, double* xreal, int n, double* inputNoise)
{
#ifdef DEBUG
  cout << "MOTestFunctions::DTLZ2()" << endl;
#endif

  if (n < 12)  // Number of decision variables, recommended value for k >= 10 (see Deb, 2002: Scalable MO Test Problems), 
  {            // hence n >= 12
    cerr << "DTLZ2: number of decision variables, n = " << n <<  ", but should be n >= 12 (then  k >= 10)! Exiting." << endl;
    exit(1);
  }	  
       
  int M = 3; // Problem dimension, i.e., M-objective
  //int k = n - M + 1; // Length of xM input vector of g(xM), i.e., xM = (xM_0,...,xM_k-1)
 
  F.assign(M, 0);

  double g = 0;
  for (int i = M; i <= n; i++)
    g += pow(((I != NULL) ? I->R[i-1] : xreal[i-1]) + ((inputNoise != NULL) ? inputNoise[i-1] : 0) - 0.5, 2);

  for (int i = 1; i <= M; i++)
  {
    double f = (1 + g);
    for (int j = M - i; j >= 1; j--)
      f *= cos((((I != NULL) ? I->R[j-1] : xreal[j-1]) + ((inputNoise != NULL) ? inputNoise[j-1] : 0)) * PI / 2);

    if (i > 1)
    {
      int index = (M - i + 1) - 1;
      f *= sin((((I != NULL) ? I->R[index] : xreal[index]) + ((inputNoise != NULL) ? inputNoise[index] : 0)) * PI / 2);
    }

    F[i-1] = f;
  }
}
