#include "WDN_MIES.h"

//------------------------------------------------------------------------------
// Class WDN_MIES
//-------------------------------------------------
// Private Members

/* Initialize Individuals in population P with uniform randomly 
 * chosen values within various variable and parameter domains
 * or read previously generated population in from file 
 * [and re-evaluate in case of robustness evaluation, then writePop() and exit]
 */
void WDN_MIES::initialize()
{
#ifdef DEBUG
  cout << "WDN_MIES::initialize()" << endl;
#endif

  MixedIntegerES::initialize();

#ifdef ROBUSTNESS_EVAL
  evaluate(); // Re-evaluate with MC samples
  writePop(-1); // When ROBUSTNESS_EVAL, RunID stored in optimalF[0] is used
  exit(0);
#endif
}

/* Determine fitness values for population P or O */
void WDN_MIES::evaluate()
{
#ifdef DEBUG
  cout << "WDN_MIES::evaluate()" << endl;
#endif

#ifdef ROBUST
  generateSampleSet();
#endif

  MixedIntegerES::evaluate();
}

/* Run simulator using the objective variables contained in Individual I */
void WDN_MIES::simulate(Individual* I) 
{ 
#ifdef DEBUG
  cout << "WDN_MIES::simulate()" << endl;
#endif

#ifdef SMSEMOA3D // Comment out for "fixed normalization" of crowding distance in NSGAII_MIES
  if (!fixedNadirIdeal)
  {
    fixedNadirIdeal = true; 
    getNadirIdeal(nadir, ideal, (n_d == 0) ? n_z : n_d);
  }
#endif
 
#ifdef ROBUST	
  WDN::robustSimulate(I, NULL, n_d); 
#else
  if (n_d == 0) // Added for integer-based optimization
  {
    I->D = I->Z;
    WDN::simulate(I, NULL, n_z, NULL);
  }
  else
    WDN::simulate(I, NULL, n_d, NULL); 
#endif    
} 

/* Create lambda offspring */
void WDN_MIES::recombineMutate()
{
#ifdef DEBUG
  cout << "WDN_MIES::recombineMutate()" << endl;
#endif

#ifdef SP
  /* Create lambda offspring by mutation of a single (recombined) parent */
  MixedIntegerES::recombineMutateSingleParent();
#else
  /* Create lambda offspring by recombination and mutation */
  MixedIntegerES::recombineMutate();
#endif
}

/* Write current population P to file */
void WDN_MIES::writePop(time_t id)
{
#ifdef DEBUG
  cout << "WDN_MIES::writePop()" << endl;
#endif

  string filename;
  fstream outfile;
  if (snapShots)
  {
    writeStep(id);

    unsigned  evalsUsed = mu + currGen * lambda;

#ifdef ROBUST   
# ifdef LHSR 
      evalsUsed = getEvalsUsed();
# else
      evalsUsed *= sampleSetSize;
# endif
#endif

    if (evalsUsed % mu != 0)
      return;
      
    filename = "RunID_" + itos((int) id) + "_P_" + itos(evalsUsed) + ".txt"; 
  }
  else
  {
#ifdef ROBUSTNESS_EVAL  
    string filename = "RunID_" + boost::lexical_cast<string>((double) optimalF[0]) + "_P_robust.txt"; 
#else
    if (currGen == 0)
      filename = "RunID_" + itos((int) id) + "_P_0.txt"; 
    else if (currGen == generations)
      filename = "RunID_" + itos((int) id) + "_P.txt"; 
    else
      return;
#endif
  }	  

  outfile.open(filename.c_str(), ios::out);

  int size;
  for (unsigned i = 0; i < mu; i++)
  {
    outfile << ((P[i]->feasible) ? "F" : "P" ) << "\t";

    size = P[i]->F.size();	  
    for (int j = 0; j < size; j++)
      outfile << P[i]->F[j] << "\t";

    outfile << "\t\t\t\t";
            
    for (unsigned j = 0; j < n_r; j++)
      outfile << P[i]->R[j] << "\t";

    for (unsigned j = 0; j < n_z; j++)
      outfile << P[i]->Z[j] << "\t";

    for (unsigned j = 0; j < n_d; j++)
      outfile << P[i]->D[j] << "\t";

    outfile << "\t\t\t\t";

    for (unsigned j = 0; j < n_sigma_r; j++)
      outfile << P[i]->S_r[j] << "\t";

    for (unsigned j = 0; j < n_sigma_z; j++)
      outfile << P[i]->S_z[j] << "\t";

    for (unsigned j = 0; j < n_prob; j++)
      outfile << P[i]->Prob[j] << "\t";

    outfile << endl << endl << endl; // Two white lines to indicate new gnuplot index
  }
  outfile.close();
}

/* Output stepsizes */
void WDN_MIES::writeStep(time_t id)
{
  string filename = "RunID_" + itos((int) id) + "_stepsizes.txt"; 
  fstream outfile(filename.c_str(), ios::out | ios::ate | ios::app);

  outfile << currGen << "\t\t";

  double mean = 0, min = DBL_MAX, max = 0;
  if (n_sigma_r > 0)
  {
    for (unsigned j = 0; j < mu; j++)
    {
      if (P[j]->S_r[0] < min)
	min = P[j]->S_r[0];

      if (P[j]->S_r[0] > max)	
	max = P[j]->S_r[0];      
      
      mean += P[j]->S_r[0];
    }
    mean /= mu;

    outfile << mean << "\t" << min << "\t" << max << "\t\t";
  }

  if (n_sigma_z > 0)
  {
    mean = 0;
    min = DBL_MAX; 
    max = 0;
    for (unsigned j = 0; j < mu; j++)
    {
      if (P[j]->S_z[0] < min)
	min = P[j]->S_z[0];

      if (P[j]->S_z[0] > max)	
	max = P[j]->S_z[0];      
      
      mean += P[j]->S_z[0];
    }
    mean /= mu;

    outfile << mean << "\t" << min << "\t" << max << "\t\t";
  }

  if (n_prob > 0)
  {
    mean = 0;
    min = DBL_MAX; 
    max = 0;
    for (unsigned j = 0; j < mu; j++)
    {
      if (P[j]->Prob[0] < min)
	min = P[j]->Prob[0];

      if (P[j]->Prob[0] > max)	
	max = P[j]->Prob[0];      
      
      mean += P[j]->Prob[0];
    }
    mean /= mu;

    outfile << (1.0/n_d + mean) << "\t" << (1.0/n_d + min) << "\t" << (1.0/n_d + max) << "\t\t";
  }

  outfile << endl;
  outfile.close();
}

/* Log time that was needed to complete generation */
void WDN_MIES::writeLog(time_t elapsed, time_t id)
{
#ifdef DEBUG
  cout << "WDN_MIES::writeLog()" << endl;
#endif

  if (snapShots)
    return;

  string filename = "RunID_" + itos((int) id) + "_log.txt"; 
  fstream outfile(filename.c_str(), ios::out | ios::ate | ios::app);

  if (currGen <= generations)
  {
    if (feedback)
      cout << endl << "Generation " << currGen << " took " << elapsed << "s" << endl;
    outfile << "Generation " << currGen << ": \t" << elapsed << "s" << endl;
  }
  else
  {
    if (feedback) 
    {
#ifdef LHSR
      cout << "Total number of evaluations used: " << getEvalsUsed() << endl;
#endif
      cout << "Optimization took " << elapsed << "s" << endl << endl;
    }
    outfile << endl << "Total optimization: \t";
#ifdef LHSR
    outfile << getEvalsUsed() << " evaluations, ";
#endif
    outfile << elapsed << "s" << endl;
  }

  outfile.close();
}
