#include "main.h"

#ifdef NSGAII
string emoa = "NSGA-II";
#else
string emoa = "SMS-EMOA";
#endif

int nreal;
int nbin;
int nnom;
int nobj;
int* obj_direction;
int ncon;
int popsize;
double pcross_real;
double pcross_bin;
double pcross_nom;
double pmut_real;
double pmut_bin;
double pmut_nom;
double eta_c;
double eta_m;
int ngen;
int nbinmut;
int nrealmut;
int nnommut;
int nbincross;
int nrealcross;
int nnomcross;
int* nbits;
double* min_realvar;
double* max_realvar;
double* min_binvar;
double* max_binvar;
int* min_nomvar;
int* max_nomvar;
int bitlength;
double *nadir; /* inserted: ereehuis */
double *ideal; /* inserted: ereehuis */
int fixedNadirIdeal; /* inserted: ereehuis */

unsigned evaluations, sampleSetSize = 1;
bool snapShots;

string ES, MOproblem;
MOTestFunctions* testFunctions = NULL;
WDN* network = NULL;

bool fileExists(string filename) 
{
  struct stat stFileInfo;

  if (stat(filename.c_str(), &stFileInfo) == 0)
    return true;

  return false;
}

/* Wrapper function to execute MOTestFunctions and WDN member functions from the Deb C nsga2-v1.1 code */
void simulatorWrapper(double* xreal, double* xbin, int** gene, int* xnom, double* obj, double* constr)
{
  if (ES == "TestFunctions_MIES")
  {
    if (MOproblem == "DTLZ1") 
      testFunctions->DTLZ1(xreal, nreal);
    else if (MOproblem == "DTLZ2")   
      testFunctions->DTLZ2(xreal, nreal);
  }
  else if (ES == "WDN_MIES")
  {
#ifdef ROBUST
    network->robustSimulate(xnom, nnom);
#else
    network->simulate(xnom, nnom);
#endif
  }

  vector<double>* F = (ES == "TestFunctions_MIES") ? testFunctions->getF() : network->getF();
  for (int i = 0; i < nobj; i++)
    obj[i] = F->at(i);
}

/* Function to allocate memory to a population */
void allocateMemoryPop(population* P, int size)
{
  P->ind = new individual[size];
  for (int i = 0; i < size; i++)
    allocateMemoryInd(&(P->ind[i]));
}

/* Function to allocate memory to an individual */
void allocateMemoryInd(individual* ind)
{
  ind->xreal = (nreal != 0) ? new double[nreal] : NULL;

  if (nbin != 0)
  {
    ind->xbin = new double[nbin];
    ind->gene = new int* [nbin];
    for (int i = 0; i < nbin; i++)
      ind->gene[i] = new int[nbits[i]];
  }
  else
  {
    ind->xbin = NULL;
    ind->gene = NULL;
  }

  ind->xnom = (nnom != 0) ? new int[nnom] : NULL;  
  ind->obj = new double[nobj];
  ind->constr = (ncon != 0) ? new double[ncon] : NULL;
}

/* Function to deallocate memory to a population */
void deallocateMemoryPop(population* P, int size)
{
  for (int i = 0; i < size; i++)
    deallocateMemoryInd(&(P->ind[i]));

  delete [] P->ind;
}

/* Function to deallocate memory to an individual */
void deallocateMemoryInd(individual* ind)
{
  delete [] ind->xreal;
  delete [] ind->xbin;
   
  for (int i = 0; i < nbin; i++)
    delete [] ind->gene[i];    
  delete [] ind->gene;

  delete [] ind->xnom;
  delete [] ind->obj;
  delete [] ind->constr;
}

/* Write current population to file */
void writePop(double id, population* P, int currGen)
{
  string filename;
  fstream outfile;
  if (snapShots)
  {
    unsigned evalsUsed; 

#ifdef NSGAII
    evalsUsed = popsize + currGen * popsize; 
#else
    evalsUsed = popsize + currGen;
#endif

#ifdef ROBUST
# ifdef LHSR
    evalsUsed = network->getEvalsUsed();
# else
    evalsUsed *= sampleSetSize;
# endif
#endif

    if (evalsUsed % popsize != 0)
      return;
      
    filename = "RunID_" + boost::lexical_cast<string>(id) + "_P_" + itos(evalsUsed) + ".txt"; 
  }
  else
  {
    if (currGen == 0)
      filename = "RunID_" + boost::lexical_cast<string>(id) + "_P_0.txt"; 
    else if (currGen == ngen)
      filename = "RunID_" + boost::lexical_cast<string>(id) + "_P.txt"; 
    else
      return;
  }	  

  outfile.open(filename.c_str(), ios::out);
  for (int i = 0; i < popsize; i++)
  {
    outfile << ((P->ind[i].constr_violation == 0.0) ? "F" : "P" ) << "\t";
 
    for (int j = 0; j < nobj; j++)
      outfile << P->ind[i].obj[j] << "\t";

    outfile << "\t\t\t\t";
            
    for (int j = 0; j < nreal; j++)
      outfile << P->ind[i].xreal[j] << "\t";

    for (int j = 0; j < nbin; j++) 
    {
      for (int m = 0; m < nbits[j]; m++)
        outfile << P->ind[i].gene[j][m] << "\t";
    }

    for (int j = 0; j < nnom; j++)
      outfile << P->ind[i].xnom[j] << "\t";

    outfile << endl << endl << endl; // Two white lines to indicate new gnuplot index
  }
  outfile.close();
}

/* Log time that was needed to complete generation */
void writeLog(double id, time_t elapsed, int currGen, bool feedback)
{
  if (snapShots)
    return;

  string filename = "RunID_" + boost::lexical_cast<string>(id) + "_log.txt"; 
  fstream outfile(filename.c_str(), ios::out | ios::ate | ios::app);

  if (currGen <= ngen)
  {
    outfile << emoa << "Generation " << currGen << ": \t" << elapsed << "s" << endl;
  }
  else
  {
#ifdef LHSR
    if (ES == "WDN_MIES")
      cout << emoa << "Total number of evaluations used: " << network.getEvalsUsed() << endl;
#endif
    cout << emoa << "Optimization took " << elapsed << "s" << endl << endl;

    outfile << endl << emoa << "Total optimization: \t";
#ifdef LHSR
    if (ES == "WDN_MIES")
      outfile << network.getEvalsUsed() << " evaluations, ";
#endif
    outfile << elapsed << "s" << endl;
  }

  outfile.close();
}

int main(int argc, char* argv[])
{ 
  if (argc < 2) 
  { 
    cerr << "No input file supplied! Exiting." << endl;	  
    return 1;
  }

#ifdef ROBUSTNESS_EVAL
  if (argc < 5)
  {
    cerr << "No pop file, RunID, and/or sample set size supplied for MC robustness evaluation! Exiting." << endl;
    return 1;
  }
#endif


#ifdef DP
  emoa += " dp";
#endif

  emoa += ": ";

  char line[LEN];
  ifstream infile(argv[1]);

  /* Read constant simulator parameters from input file */
  vector<string> constSimParams;
  skipLinesUntil(infile, HEADER_CONSTANTS);

  infile.getline(line, LEN);
  while (line[0] != '\0')
  {
    constSimParams.push_back(strtok(line, " ;"));
    findAndReplace(constSimParams.back(), SPACE, " ");
    infile.getline(line, LEN);
  }

  /* Read variable simulator parameter names from input file */	
  vector<string> varSimParamNames;
  skipLinesUntil(infile, HEADER_NAMES);

  infile.getline(line, LEN);
  while (line[0] != '\0')
  {
    varSimParamNames.push_back(strtok(line, " ;"));
    infile.getline(line, LEN);
  }

  /* Read MIES parameters from input file */
  skipLinesUntil(infile, HEADER_MIES);
  
  unsigned n_r       = atoi(getValue(infile)),
           n_z       = atoi(getValue(infile)),
           n_d       = atoi(getValue(infile)),
           n_sigma_r = atoi(getValue(infile)),
           n_sigma_z = atoi(getValue(infile)),
           n_prob    = atoi(getValue(infile)),
           n_f       = atoi(getValue(infile)),
	   mu        = atoi(getValue(infile)),
	   rho       = atoi(getValue(infile)),
           kappa     = atoi(getValue(infile)),
           lambda    = atoi(getValue(infile));

  vector<double> lBound_r, uBound_r; 
  vector<int> lBound_z, uBound_z, lBound_d, uBound_d; 

  getBound(infile, &lBound_r, NULL, n_r);
  getBound(infile, &uBound_r, NULL, n_r);

  getBound(infile, NULL, &lBound_z, n_z);
  getBound(infile, NULL, &uBound_z, n_z);

  getBound(infile, NULL, &lBound_d, n_d);
  getBound(infile, NULL, &uBound_d, n_d);

  double prob_min = atof(getValue(infile)),
	 prob_max = atof(getValue(infile));

  bool disRec_r       = (atoi(getValue(infile)) == 0) ? false : true,
       disRec_z       = (atoi(getValue(infile)) == 0) ? false : true, 
       disRec_sigma_r = (atoi(getValue(infile)) == 0) ? false : true,
       disRec_sigma_z = (atoi(getValue(infile)) == 0) ? false : true, 
       disRec_prob    = (atoi(getValue(infile)) == 0) ? false : true,
       selfAdaptation = (atoi(getValue(infile)) == 0) ? false : true; 

  vector<int> direction; 
  for (unsigned i = 0; i < n_f; i++)
    direction.push_back(atoi(getValue(infile)));

  unsigned selectDimension = atoi(getValue(infile));

  vector<unsigned> selectFunction;
  for (unsigned i = 0; i < selectDimension; i++)
    selectFunction.push_back(atoi(getValue(infile)));

  evaluations = atoi(getValue(infile));

  vector<double> optimalF;
  for (unsigned i = 0; i < n_f; i++)
    optimalF.push_back(atof(getValue(infile)));

  double initialSigma_r = atof(getValue(infile)),
	 initialSigma_z = atof(getValue(infile)),
         initialProb    = atof(getValue(infile));

  string initialPopFileName = getStringValue(infile);
#ifdef ROBUSTNESS_EVAL
  initialPopFileName = argv[2];

  optimalF[0] = atof(argv[3]); // Use optimalF[0] for storing RunID
#endif  

  int randomSeed = atoi(getValue(infile));

#ifdef MIES
  int feedback = (atoi(getValue(infile)) == 1) ? true : false;
#else
  int feedback = (atoi(getValue(infile)) == 0) ? false : true;
#endif

  ES = getStringValue(infile); 

  if (ES == "TestFunctions_MIES")
    MOproblem = getStringValue(infile);

  bool excludeParent = (atoi(getValue(infile)) == 0) ? false : true;

  double cvDemand, cvRoughness;
  if (ES == "WDN_MIES")
  {
    cvDemand      = atof(getValue(infile));
    cvRoughness   = atof(getValue(infile));
    sampleSetSize = atoi(getValue(infile));

#ifdef ROBUSTNESS_EVAL
    sampleSetSize = atoi(argv[4]);
#endif
  }

  /* Read simulator file names from input file */
  skipLinesUntil(infile, HEADER_SIMULATOR);
  string simulator         = getStringValue(infile),
	 path              = getStringValue(infile),
	 simInputFileName  = getStringValue(infile),
	 simOutputFileName = getStringValue(infile);
  
  unsigned timeOutValue = 0;

  infile.close();

  findAndReplace(initialPopFileName, SPACE, " ");
  findAndReplace(MOproblem, SPACE, " ");
  findAndReplace(simulator, SPACE, " ");
  findAndReplace(path, SPACE, " ");
  findAndReplace(simInputFileName, SPACE, " ");
  findAndReplace(simOutputFileName, SPACE, " ");

  ifstream* initialPopFile = (initialPopFileName == " ") ? NULL : new ifstream(initialPopFileName.c_str());

  simulator = "\"\"\"" + simulator + "\" \"" + path + "\"\"\"";  

  snapShots = fileExists(SNAPSHOTS);

#ifdef MIES 
  if (ES == "TestFunctions_MIES")
  {
    TestFunctions_MIES testMies(n_r, n_z, n_d, n_sigma_r, n_sigma_z, n_prob, n_f, 
		                 mu, rho, kappa, lambda,  
		                   lBound_r, uBound_r, lBound_z, uBound_z, 
		                     lBound_d, uBound_d, prob_min,  prob_max,
			               disRec_r, disRec_z, disRec_sigma_r, disRec_sigma_z, disRec_prob,
			                 selfAdaptation, direction, 
			                   evaluations, optimalF, initialSigma_r, initialSigma_z, initialProb, initialPopFile, 
					     randomSeed, feedback, timeOutValue,
				               selectDimension, selectFunction,
					         MOproblem, excludeParent);
    testMies.run();
  }
  else if (ES == "WDN_MIES")
  {
    WDN_MIES wdnMies(n_r, n_z, n_d, n_sigma_r, n_sigma_z, n_prob, n_f, mu, rho, kappa, lambda,  
		       lBound_r, uBound_r, lBound_z, uBound_z, lBound_d, uBound_d, prob_min, prob_max,
	                 disRec_r, disRec_z, disRec_sigma_r, disRec_sigma_z, disRec_prob,
		           selfAdaptation, direction, evaluations, optimalF, 
                             initialSigma_r, initialSigma_z, initialProb,
      		               initialPopFile, randomSeed,  feedback, snapShots,
			         timeOutValue,
                                   simInputFileName, simOutputFileName, selectDimension, selectFunction,
			             excludeParent,
				       cvDemand, cvRoughness, sampleSetSize
#ifdef ROBUSTNESS_EVAL
                                       , false // Use Monte Carlo sampling for robustness evaluation
#endif
				       ); 
    wdnMies.run();
  }
#else
  if (ES == "TestFunctions_MIES")
    testFunctions = new MOTestFunctions;
  else if (ES == "WDN_MIES")
  {
    network = new WDN(simInputFileName, simOutputFileName, cvDemand, cvRoughness, sampleSetSize);  

    if (network->isParallelExp())
    {
      /* Enable selection of zero diameter */ 
      for (unsigned i = 0; i < n_d; i++)
	uBound_d[i]++;
    }
  }

  /* NSGA-II execution */
  nreal = n_r;
  nbin = n_z;
  nnom = n_d;
  nobj = selectDimension;
  ncon = 0;
  popsize = mu;
  pcross_real = P_CROSS_R;
  pcross_bin = P_CROSS_B;
  pcross_nom = P_CROSS_N;
  pmut_real = 1/(double) nreal;
  pmut_bin = 1/(double) nbin;
  pmut_nom = 1/(double) nnom;
  eta_c = DIST_IDX_C;
  eta_m = DIST_IDX_M;
  nbinmut = 0;
  nrealmut = 0;
  nnommut = 0;
  nbincross = 0;
  nrealcross = 0;
  nnomcross = 0;
  fixedNadirIdeal = 0;

  population* P = new population,
	    * O = new population,
	    * Q = new population;

  /* Use clock or supplied randomSeed for uniReal() */
  randGenerator.seed(static_cast<unsigned int>((randomSeed == -1) ? time(NULL) : randomSeed));
  seed = uniReal() * 0.9 + 0.05; // seed must be in ]0,1[

  if (popsize < 4 || popsize % 4 != 0)
  {
    cerr << "Population size should be a multiple of 4! Exiting." << endl;
    return 1;
  }

  /* Determine number of generations possible with current parameter settings */
#ifndef ROBUST
  sampleSetSize = 1;
#endif

  if (evaluations > 0)
#ifdef NSGAII
    ngen = (evaluations - popsize * sampleSetSize) / (popsize * sampleSetSize); 
#else
    ngen = (evaluations - popsize * sampleSetSize) / sampleSetSize;
#endif
  else
  {
    cerr << "Supplied value for mumber of evaluations is " << evaluations << ", but should be larger than 0! Exiting." << endl;
    exit(1);
  }

  obj_direction = (int*) malloc(nobj * sizeof(int)); 
  for (int i = 0; i < nobj; i++) 
    obj_direction[i] = direction[i];

  if (nreal != 0)
  {
    min_realvar = (double*) malloc(nreal * sizeof(double));
    max_realvar = (double*) malloc(nreal * sizeof(double));
    
    for (int i = 0; i < nreal; i++)
    {
      min_realvar[i] = lBound_r[i];       
      max_realvar[i] = uBound_r[i];
    }
  }

  if (nbin != 0) // TODO: for simulation, only use integer part (discard fraction part)
  {
    nbits = (int*) malloc(nbin * sizeof(int));
    min_binvar = (double*) malloc(nbin * sizeof(double));
    max_binvar = (double*) malloc(nbin * sizeof(double));
    bitlength = 0;
        
    for (int i = 0; i < nbin; i++)
    {
      min_binvar[i] = lBound_z[i];
      max_binvar[i] = uBound_z[i];

      nbits[i] = (int) ceil(pow(uBound_z[i] - lBound_z[i], 0.5)); 
      bitlength += nbits[i];
    }
  }

  if (nnom != 0)
  {
    min_nomvar = (int*) malloc(nnom * sizeof(int));
    max_nomvar = (int*) malloc(nnom * sizeof(int));
    
    for (int i = 0; i < nnom; i++)
    {
      min_nomvar[i] = lBound_d[i]; 
      max_nomvar[i] = uBound_d[i];
    }
  }

  test_problem = &simulatorWrapper;

  allocateMemoryPop(P, popsize);

#ifdef NSGAII
  allocateMemoryPop(O, popsize);
  allocateMemoryPop(Q, 2*popsize);
#else
  individual* q = new individual;
  allocateMemoryInd(q);
#endif

  if (ES == "WDN_MIES")
  {
    fixedNadirIdeal = 0; // Set to 1 for "fixed normalization" of crowding distance in nsga2-v1.1 (basic NSGA-II)
    
    vector<double> nadir_, ideal_;
    network->getNadirIdeal(nadir_, ideal_, nnom);

    nadir = new double[3];
    ideal = new double[3];
    for (int i = 0; i < 3; i++)
    {    
      nadir[i] = nadir_[i];
      ideal[i] = ideal_[i];
    }

#ifdef SMSEMOA3D
    setNadirIdeal(nadir, ideal);
#endif
  }

  time_t start = time(NULL); 
  time_t startGen = start;

  randomize();

  if (feedback)
    cout << endl << emoa << "Generating initial generation" << endl;

#ifdef ROBUST  
  if (ES == "WDN_MIES")
    network->generateSampleSet();
#endif
  initialize_pop(P);
  decode_pop(P);
  evaluate_pop(P);
#ifdef NSGAII  
  assign_rank_and_crowding_distance(P);
#endif

  writePop(seed, P, 0);
  time_t elapsed = time(NULL) - startGen;
  writeLog(seed, elapsed, 0, feedback);

  for (int i = 1; i <= ngen; i++)
  {
    startGen = time(NULL);

    if (feedback && (i % REPORT_WINDOW == 1 || i == ngen)) 
      cout << endl << emoa << "At generation " << i << "/" << ngen << endl;

#ifdef ROBUST  
    if (ES == "WDN_MIES")
      network->generateSampleSet();
#endif

#ifdef NSGAII 
    /* NSGA-II specific part of evolution loop */
    selection(P, O);
    mutation_pop(O);
    decode_pop(O);
    evaluate_pop(O);
    merge(P, O, Q);
    fill_nondominated_sort(Q, P);
#else
    /* 3D SMS-EMOA specific part of evolution loop */
    generate(P, q);
    reduce(P, q);
#endif

    writePop(seed, P, i);
    elapsed = time(NULL) - startGen;
    writeLog(seed, elapsed, i, feedback);

#ifdef LHSR
    if (ES == "WDN_MIES")
      ngen = network->updateTotalGen(ngen, currGen, evaluations, lambda);
#endif
  }

  elapsed = time(NULL) - start;
  writeLog(seed, elapsed, ngen+1, feedback);

  if (ES == "WDN_MIES")
  {
    cout << "IdealTP: " << endl;
    for (int i = 0; i < 3; i++)
      cout << "F" << i << " = " << ideal[i] << endl;
    cout << endl;

    cout << "BoundTP: " << endl;
    for (int i = 0; i < 3; i++)
      cout << "F" << i << " = " << nadir[i] << endl;
    cout << endl;
    
    delete [] nadir;
    delete [] ideal;
  }  

  if (nreal != 0)
  {
    free(min_realvar); 
    free(max_realvar);
  }
  
  if (nbin != 0) 
  {
    free(min_binvar);
    free(max_binvar);
    free(nbits);
  }

  if (nnom != 0)
  {
    free(min_nomvar); 
    free(max_nomvar);
  }
  
  deallocateMemoryPop(P, popsize);
  delete P;

#ifdef NSGAII
  deallocateMemoryPop(O, popsize);
  delete O;
  deallocateMemoryPop(Q, 2*popsize);
  delete Q;
#else
  deallocateMemoryInd(q);
  delete q;
#endif

  delete testFunctions;
  delete network;
#endif

  return 0;
}

