#include "MixedIntegerES.h"

//------------------------------------------------------------------------------
// Class MixedIntegerES
//-------------------------------------------------
// Public Members

MixedIntegerES::MixedIntegerES(unsigned n_r_, unsigned n_z_, unsigned n_d_, unsigned n_sigma_r_, unsigned n_sigma_z_, unsigned n_prob_, unsigned n_f_, 
		                unsigned mu_, unsigned rho_, unsigned kappa_, unsigned lambda_, 
		              	  vector<double>& lBound_r_, vector<double>& uBound_r_, vector<int>& lBound_z_, vector<int>& uBound_z_, 
		                    vector<int>& lBound_d_, vector<int>& uBound_d_, double prob_min_, double prob_max_,
			              bool disRec_r_, bool disRec_z_, bool disRec_sigma_r_, bool disRec_sigma_z_, bool disRec_prob_,
			                bool selfAdaptation_, vector<int>& direction_, 
			                  unsigned evaluations_, vector<double>& optimalF_, 
					    double initialSigma_r_, double initialSigma_z_, double initialProb_,
					      ifstream* initialPopFile_, int randomSeed_, bool feedback_) 
{
  n_r = n_r_; n_z = n_z_; n_d = n_d_; n_sigma_r = n_sigma_r_; n_sigma_z = n_sigma_z_; n_prob = n_prob_; n_f = n_f_; 
  mu = mu_; rho = rho_; kappa = kappa_; lambda = lambda_;
  lBound_r = lBound_r_; uBound_r = uBound_r_; lBound_z = lBound_z_; uBound_z = uBound_z_; lBound_d = lBound_d_; uBound_d = uBound_d_; 
  prob_min = prob_min_; prob_max = prob_max_;
  disRec_r = disRec_r_; disRec_z = disRec_z_; disRec_sigma_r = disRec_sigma_r; disRec_sigma_z = disRec_sigma_z_; disRec_prob = disRec_prob_; 
  selfAdaptation = selfAdaptation_; direction = direction_; 
  evaluations = evaluations_; optimalF = optimalF_; 
  initialSigma_r = initialSigma_r_; initialSigma_z = initialSigma_z_; initialProb = initialProb_;
  initialPopFile = initialPopFile_; randomSeed = randomSeed_; feedback = feedback_;

  debug = false;

  /* Insert direction and optimalF value for product/sum of (optionally normalized) F values */
  if (direction[0] == 1) // Product in case of F values to be maximized
  {
    direction.push_back(1); // Product->MAX

    optimalF.push_back(1);
    for (unsigned i = 0; i < n_f; i++)
      optimalF[n_f] *= optimalF[i];
  }
  else                   // Sum in case of F values to be minimized
  {
    direction.push_back(-1); // Sum->min

    optimalF.push_back(0);
    for (unsigned i = 0; i < n_f; i++)
      optimalF[n_f] += optimalF[i];
  }

  P.assign(mu, NULL);
  O.assign(lambda, NULL);
  bestF.resize(n_f+1); // Reserve space for product/sum of (optionally normalized) F values as well

  /* Determine number of generations possible with current parameter settings */
  if (evaluations > 0)
    generations = (evaluations - ((initialPopFile == NULL) ? mu : 0)) / lambda; 
  else
    generations = USHRT_MAX; // OptimalF supplied, convergence behavoir study 

  currGen = 0;

  /* In case of no self-adaptation, set maximum stepsize values to (fraction of which) stepsizes will be constrained */
  if (!selfAdaptation)
  {
    /* Maximum stepsize values for real-valued variables */
    maxSigma_r.assign(n_sigma_r, 0);

    for (unsigned i = 0; i < n_sigma_r; i++)
    {
      if (i == n_sigma_r - 1) // maxSigma_r for last S_r:
      {
        for (unsigned j = i; j < n_r; j++) // Take equal to the average rough distance to optimum (=delta) of variables R[i...n_r-1]
          maxSigma_r[i] += uBound_r[j] - lBound_r[j];

        maxSigma_r[i] /= n_r - n_sigma_r + 1;
        maxSigma_r[i] /= sqrt(n_r);
      }
      else // Take maxSigma_r equal to rough distance to optimum (=delta) of variable R[i] 
        maxSigma_r[i] = (uBound_r[i] - lBound_r[i]) / sqrt(n_r);
    }

    /* Maximum stepsize values for integer variables */
    maxSigma_z.assign(n_sigma_z, 0);

    for (unsigned i = 0; i < n_sigma_z; i++)
    {
      if (i == n_sigma_z - 1) // maxSigma_z for last S_z:
      {
        for (unsigned j = i; j < n_z; j++) // Take equal to the average rough distance to optimum (=delta) of variables Z[i...n_z-1]
          maxSigma_z[i] += uBound_z[j] - lBound_z[j];

        maxSigma_z[i] /= n_z - n_sigma_z + 1;
        maxSigma_z[i] /= sqrt(n_z);
      }
      else // Take maxSigma_z equal to rough distance to optimum (=delta) of variable Z[i] 
        maxSigma_z[i] = (uBound_z[i] - lBound_z[i]) / sqrt(n_z);
    }
  }
}

MixedIntegerES::~MixedIntegerES()
{
#ifdef DEBUG
  cout << "MixedIntegerES::~MixedIntegerES()" << endl;
#endif

  for (unsigned i = 0; i < mu; i++)
    delete P[i];

  for (unsigned i = 0; i < lambda; i++)
    delete O[i];
}

/* Execute the MIES */
void MixedIntegerES::run()
{
#ifdef DEBUG
  cout << "MixedIntegerES::run()" << endl;
#endif

  time_t start = time(NULL); 
  time_t startGen = start;

  /* Random seed: if -1 use system clock, else use supplied seed */
  id = (randomSeed == -1) ? start : randomSeed; 

  /* Initialize random number generator */
  randGenerator.seed(static_cast<unsigned int>(id));

  /* Generate and evaluate initial population or read in from file */
  initialize();

  /* Write current population to file */
  writePop(id);

  time_t elapsed = time(NULL) - startGen;
  writeLog(elapsed, id);
 
  for (currGen = 1; currGen <= generations; currGen++)
  {
    /* Print debug info in case of slow convergence */	
    if (currGen > DEBUG_FROM_GEN)
      debug = true;

    /* Log time needed to complete generation */	  
    startGen = time(NULL);

    increaseAge();

    recombineMutate();

    if (feedback)
      cout << endl << "Evaluating generation " << currGen << "/" << generations << endl;

    evaluate();

    select();

    /* Write current population to file */
    writePop(id);

    elapsed = time(NULL) - startGen;
    writeLog(elapsed, id);

#ifdef LHSR
    updateTotalGen(generations, currGen, evaluations, lambda);
#endif

    if (optimumReached())
    {
      cout << endl << "Optimum reached in generation " << currGen << endl;
      return;
    }
  } 
  
  /* Optimization complete: print total elapsed time, and best objective value(s) */
  cout << endl;

  elapsed = time(NULL) - start;
  writeLog(elapsed, id);

  for (unsigned i = 0; i < n_f+1; i++)
  {
    if (direction[i] == 1)
    {
      if (bestF[i] == 0)
        continue;
    }	    
    else
    {
      if (bestF[i] == DBL_MAX)
        continue;
    } 

    if (i == n_f)
      cout << "Optimal " << ((direction[i] == 1) ? "product" : "sum") << " of F values = " << bestF[i] << endl;
    else
      cout << "bestF[" << i << "] = " << bestF[i] << endl;      
  }
}

//-------------------------------------------------
// Protected Members

/* Initialize Individuals in population P with uniform randomly 
 * chosen values within various variable and parameter domains
 * or read previously generated population in from file
 */
void MixedIntegerES::initialize()
{
#ifdef DEBUG
  cout << "MixedIntegerES::initialize()" << endl;
#endif

  /* Generate and evaluate if initial population not supplied in file */
  if (initialPopFile == NULL)
  {	     
    if (feedback)
      cout << "Generating initial population" << endl;
     
    for (unsigned i = 0; i < mu; i++) 
      P[i] = new Individual(n_r, n_z, n_d, n_sigma_r, n_sigma_z, n_prob, n_f,
			     lBound_r, uBound_r, lBound_z, uBound_z, lBound_d, uBound_d, prob_min, prob_max,
			       initialSigma_r, initialSigma_z, initialProb);

    if (feedback)
      cout << "Evaluating initial population" << endl;

    evaluate(); // currGen = 0
  }
  else
  {
    if (feedback)
      cout << "Loading initial population from file" << endl;

    /* Reset best fitness values */
    for (unsigned i = 0; i < n_f+1; i++)
    {
      if (direction[i] == 1)
        bestF[i] = 0;
      else
        bestF[i] = DBL_MAX;
    }

    vector<double> R, S_r, S_z, Prob, F;
    vector<int> Z, D;
    bool feasible;
    char line[LEN], * next;

    for (unsigned i = 0; i < mu; i++) 
    {
      R.assign(n_r, 0);
      Z.assign(n_z, 0);
      D.assign(n_d, 0);

      S_r.assign(n_sigma_r, 0);
      S_z.assign(n_sigma_z, 0);
      Prob.assign(n_prob, 0);

      F.assign(3*n_f+1, 0); 

      /* Read in Individual */

      feasible = (initialPopFile->get() == 'F') ? true : false;

      initialPopFile->getline(line, LEN);

      /* Fitness values */
      F[0] = strtod(line, &next);
#ifdef MASPROJ
      for (unsigned j = 1; j < 3*n_f+1; j++)
#else
      for (unsigned j = 1; j < n_f; j++)
#endif
        F[j] = strtod(next, &next);

      /* Object variables */
      for (unsigned j = 0; j < n_r; j++)
	      R[j] = strtod(next, &next);

      for (unsigned j = 0; j < n_z; j++)
	      Z[j] = (int) strtod(next, &next);

      for (unsigned j = 0; j < n_d; j++)
	      D[j] = (int) strtod(next, &next);

      /* Strategy parameters */
      for (unsigned j = 0; j < n_sigma_r; j++)
	      S_r[j] = strtod(next, &next);

      for (unsigned j = 0; j < n_sigma_z; j++)
	      S_z[j] = strtod(next, &next);

      for (unsigned j = 0; j < n_prob; j++)
	      Prob[j] = strtod(next, &next);

      P[i] = new Individual(R, Z, D, S_r, S_z, Prob, F, feasible);

      /* Skip two white lines */
      initialPopFile->getline(line, LEN);
      initialPopFile->getline(line, LEN);
    }

    initialPopFile->close();
  }
}

/* Determine fitness values for population P or O */
void MixedIntegerES::evaluate()
{
#ifdef DEBUG
  cout << "MixedIntegerES::evaluate()" << endl;
#endif

  /* Reset best fitness values */
  for (unsigned i = 0; i < n_f+1; i++)
  {
    if (direction[i] == 1)
      bestF[i] = 0;
    else
      bestF[i] = DBL_MAX;
  }

  if (currGen == 0)
  {
    for (unsigned i = 0; i < mu; i++)
    {
      if (feedback)
	      cout << endl << "Running simulator for generation " << currGen 
               << "/" << generations << ", Individual " << i+1 << "/" << mu << endl << endl;
      simulate(P[i]);
    }
  }
  else
  {
    for (unsigned i = 0; i < lambda; i++)
    {
      if (feedback)
	      cout << endl << "Running simulator for generation " << currGen 
               << "/" << generations << ", Individual " << i+1 << "/" << lambda << endl << endl;
      simulate(O[i]);
    }
  }
}

/* Increase age of Individuals in population P by one */
void MixedIntegerES::increaseAge()
{
  for (unsigned i = 0; i < mu; i++)
    P[i]->age++;
}

/* Create lambda offspring by recombination and mutation */
void MixedIntegerES::recombineMutate() 
{
#ifdef DEBUG
  cout << "MixedIntegerES::recombineMutate()" << endl;
#endif

  vector<int> parents;

  /* Generate lambda offspring */
  for (unsigned i = 0; i < lambda; i++)
  {
    /* Draw uniform randomly rho parent indices from population P */
    for (unsigned j = 0; j < rho; j++) 
      parents.push_back(uniInt() % mu);

    O[i] = new Individual();
 
    /* Recombine object variables and strategy parameters */
    applyRecomb(O[i], parents);

    /* Mutate strategy parameters and object variables */
    applyMut(O[i]);

    /* Fitness values */
    O[i]->F.assign(3*n_f+1, 0); // Reserve space for, in case of
                                // A posteriori method: normalized+penalized F values, (sum/product), normalized F values, original F values 
                                // A priori method:     (norm+pen), sum/product of normalized F values, normalized F values, original F values
		                // (see SingleAisle_MIES)  
    parents.clear();
  }
}

/* Create lambda offspring by mutation of a single (recombined) parent */
void MixedIntegerES::recombineMutateSingleParent() 
{
#ifdef DEBUG
  cout << "MixedIntegerES::recombineMutateSingleParent()" << endl;
#endif

  Individual* singleParent;

  /* Draw random parent from population P to use as singleParent, and place at position 0
   * so it can be excluded in comparison score calculation, see NSGAII_Steady_MIES) */
  if (rho == 1)
  {
    int rand = uniInt() % mu;
    Individual* temp = P[0];
    P[0] = P[rand];
    P[rand] = temp;
    
    singleParent = P[0];
  }
  /* Draw uniform randomly rho parent indices from population P and recombine into singleParent */
  else
  {
    vector<int> parents;
    for (unsigned j = 0; j < rho; j++) 
      parents.push_back(uniInt() % mu);

    singleParent = new Individual();
    applyRecomb(singleParent, parents);
  }

  /* Generate lambda offspring */
  for (unsigned i = 0; i < lambda; i++)
  { 
    O[i] = new Individual(singleParent);

    /* Mutate strategy parameters and object variables */
    applyMut(O[i]);

    /* Fitness values */
    O[i]->F.assign(3*n_f+1, 0); // Reserve space for, in case of
                                // A posteriori method: normalized+penalized F values, (sum/product), normalized F values, original F values 
                                // A priori method:     (norm+pen), sum/product of normalized F values, normalized F values, original F values
		                            // (see SingleAisle_MIES)  
  }

  if (rho > 1)
    delete singleParent;
}

void MixedIntegerES::applyRecomb(Individual* I, vector<int>& parents)
{
#ifdef DEBUG
  cout << "MixedIntegerES::applyRecomb()" << endl;
#endif

  /* Real-valued */
  if (n_r > 0) 
  {
    /* Real-valued: object variables */
    I->R.assign(n_r, 0);

    if (disRec_r) // Discrete recombination
    {
      for (unsigned i = 0; i < n_r; i++) // Per variable select the value of one of the rho parents
        I->R[i] = P[parents[uniInt() % rho]]->R[i];
    } 
    else // Intermediary recombination
    {
      for (unsigned i = 0; i < n_r; i++) // Per variable:
      {
        for (unsigned j = 0; j < rho; j++) // 1) add values of the rho parents
          I->R[i] += P[parents[j]]->R[i];

        I->R[i] /= rho;                    // 2) divide by the number of parents
      }
    }

    /* Real-valued: strategy parameters */
    I->S_r.assign(n_sigma_r, 0);

    if (disRec_sigma_r) // Discrete recombination
    {
      for (unsigned i = 0; i < n_sigma_r; i++) // Per parameter the select value of one of the rho parents
        I->S_r[i] = P[parents[uniInt() % rho]]->S_r[i];
    } 
    else // Intermediary recombination
    {
      for (unsigned i = 0; i < n_sigma_r; i++) // Per parameter:
      {
        for (unsigned j = 0; j < rho; j++) // 1) add values of the rho parents
          I->S_r[i] += P[parents[j]]->S_r[i];

        I->S_r[i] /= rho;                  // 2) divide by the number of parents
      }
    }
  }
  
  /* Integer */
  if (n_z > 0) 
  {
    /* Integer: object variables */
    I->Z.assign(n_z, 0);

    if (disRec_z) // Discrete recombination
    {
      for (unsigned i = 0; i < n_z; i++) // Per variable select the value of one of the rho parents
        I->Z[i] = P[parents[uniInt() % rho]]->Z[i];
    } 
    else // Intermediary recombination
    {
      for (unsigned i = 0; i < n_z; i++) // Per variable:
      {
        for (unsigned j = 0; j < rho; j++)        // 1) add values of the rho parents
          I->Z[i] += P[parents[j]]->Z[i];

        I->Z[i] = lround((double) I->Z[i] / rho); // 2) divide by the number of parents
      }
    }

    /* Integer: strategy parameters */
    I->S_z.assign(n_sigma_z, 0);

    if (disRec_sigma_z) // Discrete recombination
    {
      for (unsigned i = 0; i < n_sigma_z; i++) // Per parameter select the value of one of the rho parents
        I->S_z[i] = P[parents[uniInt() % rho]]->S_z[i];
    } 
    else // Intermediary recombination
    {
      for (unsigned i = 0; i < n_sigma_z; i++) // Per parameter:
      {
        for (unsigned j = 0; j < rho; j++) // 1) add values of the rho parents
          I->S_z[i] += P[parents[j]]->S_z[i];

        I->S_z[i] /= rho;                  // 2) divide by the number of parents
      }
    }
  }

  /* Discrete */
  if (n_d > 0) 
  {  
    /* Discrete: object variables */
    I->D.assign(n_d, 0);

    for (unsigned i = 0; i < n_d; i++) // Discrete recombination: per variable select the value of one of the rho parents
      I->D[i] = P[parents[uniInt() % rho]]->D[i];
   
    /* Discrete: strategy parameters */
    I->Prob.assign(n_prob, 0);

    if (disRec_prob) // Discrete recombination
    {
      for (unsigned i = 0; i < n_prob; i++) // Per parameter select the value of one of the rho parents
        I->Prob[i] = P[parents[uniInt() % rho]]->Prob[i];
    } 
    else // Intermediary recombination
    {
      for (unsigned i = 0; i < n_prob; i++) // Per parameter:
      {
        for (unsigned j = 0; j < rho; j++) // 1) add values of the rho parents
          I->Prob[i] += P[parents[j]]->Prob[i];

        I->Prob[i] /= rho;                 // 2) divide by the number of parents
      }
    }
  }
}

void MixedIntegerES::applyMut(Individual* I) 
{
#ifdef DEBUG
  cout << "MixedIntegerES::applyMut()" << endl;
#endif

  double N_c;
  double tau, tau1;
  double epsilon;
 
  /* 1: Mutation of real-valued variables */
  N_c  = normReal();
  tau  = 1.0/sqrt(2.0*n_r);
  tau1 = 1.0/sqrt(2.0*sqrt(n_r));
  epsilon = 1e-30;

  /* Real-valued: strategy parameters */
  if (n_sigma_r == 1)
  {
    I->S_r[0] = max(epsilon, I->S_r[0] * exp(tau*N_c));

    if (!selfAdaptation) // Fixed convergence of real-vars stepsize
      I->S_r[0] = fixedConvergence(I->S_r[0], epsilon, maxSigma_r[0]-epsilon);  
  }
  else
  {
    for (unsigned i = 0; i < n_sigma_r; i++)
      I->S_r[i] = max(epsilon, I->S_r[i] * exp(tau*N_c + tau1*normReal()));

    if (!selfAdaptation) // Fixed convergence of real-vars stepsizes
    {
      for (unsigned i = 0; i < n_sigma_r; i++)
        I->S_r[i] = fixedConvergence(I->S_r[i], epsilon, maxSigma_r[i]-epsilon);  
    }
  }

  /* Real-valued: object variables */
  for (unsigned i = 0; i < n_r; i++)
  {
    I->R[i] += I->S_r[min(n_sigma_r-1, i)] * normReal();
    
    /* Boundary rules for keeping variables within bounds */
    I->R[i] = transformR(I->R[i], lBound_r[i], uBound_r[i]);
  }

  /* 2: Mutation of integer variables */
  N_c  = normReal();
  tau  = 1.0/sqrt(2.0*n_z);
  tau1 = 1.0/sqrt(2.0*sqrt(n_z));

  /* Integer: strategy parameters */
  if (n_sigma_z == 1)
  {
    I->S_z[0] = max(1.0, I->S_z[0] * exp(tau*N_c)); 

    if (!selfAdaptation) // Fixed convergence of integer-vars stepsize
      I->S_z[0] = fixedConvergence(I->S_z[0], 1.0, maxSigma_z[0]-1.0); 
  }
  else
  {
    for (unsigned i = 0; i < n_sigma_z; i++)
      I->S_z[i] = max(1.0, I->S_z[i] * exp(tau*N_c + tau1*normReal()));

    if (!selfAdaptation) // Fixed convergence of integer-vars stepsizes
    {
      for (unsigned i = 0; i < n_sigma_z; i++)
        I->S_z[i] = fixedConvergence(I->S_z[i], 1.0, maxSigma_z[i]-1.0);  
    }
  }

  /* Integer: object variables */
  double u1, u2, p, s;
  int G1, G2;
  for (unsigned i = 0; i < n_z; i++)
  {
    u1 = uniReal();
    u2 = uniReal();
    s = I->S_z[min(n_sigma_z-1, i)];

    p = 1.0 - (s/n_z) / (1.0 + sqrt(1.0+pow(s/n_z, 2.0))); 
    G1 = (int) floor(log(1.0-u1) / log(1.0-p));
    G2 = (int) floor(log(1.0-u2) / log(1.0-p));

    I->Z[i] += G1 - G2;

    /* Boundary rules for keeping variables within bounds */
    I->Z[i] = transformZ(I->Z[i], lBound_z[i], uBound_z[i]);
  } 

  /* 3: Mutation of discrete variables */
  N_c  = normReal();
  double gamma = 0.4;
  tau  = 1.0/sqrt(2.0*n_d); 
  tau1 = 1.0/sqrt(2.0*sqrt(n_d));
  
  /* Discrete: strategy parameters */
  if (n_prob == 1)
  {
    I->Prob[0] = 1.0 / (1.0 + ((1.0-I->Prob[0]) / I->Prob[0]) * exp(-gamma*N_c)); 

    if (selfAdaptation) // Boundary rules for keeping probabilities within bounds
    {
      if (I->Prob[0] > prob_max - 1.0/n_d) 
        I->Prob[0] = prob_max - 1.0/n_d;
    }
    else                // Fixed convergence of probability
      I->Prob[0] = fixedConvergence(I->Prob[0], prob_min, prob_max);
  }
  else // TODO: fix multiple stepsize case (like single probability case)
  {
    for (unsigned i = 0; i < n_prob; i++)
    {
      I->Prob[i] = 1.0 / (1.0 + ((1.0-I->Prob[i]) / I->Prob[i]) * exp(-tau*N_c - tau1*normReal()));

      if (selfAdaptation) // Boundary rules for keeping probabilities within bounds
        I->Prob[i] = transformR(I->Prob[i], prob_min, prob_max);
      else                // Fixed convergence of probabilities
        I->Prob[i] = fixedConvergence(I->Prob[i], prob_min, prob_max);
    }
  } 

  /* Discrete: object variables */
  int D_old;
  for (unsigned i = 0; i < n_d; i++)
  {
    if (uniReal() < (1.0/n_d + I->Prob[min( n_prob-1, i)])) 
    {
      D_old = I->D[i];

      do 
      {
        I->D[i] = uniInt() % (uBound_d[i] - lBound_d[i] + 1) + lBound_d[i];
      }
      while (I->D[i] == D_old);
    }
  }
}

/* Reflect double x back into domain [a,b]; b must be strictly greater than a */
double MixedIntegerES::transformR(double x, double a, double b)
{
  double y, y1, x1;

  y = (x-a) / (b-a);

  if ((int) floor(y) % 2 == 0)
    y1 = abs(y - floor(y));
  else
    y1 = 1 - abs(y - floor(y));

  x1 = a + (b-a)*y1;
  
#ifdef DEBUG
  cout << "MixedIntegerES::transformR(): x=" << x << ", a=" << a << ", b=" << b << ", x1=" << x1 << endl; 
#endif

  return x1;
}

/* Reflect integer x back into domain [a,b]; b must be strictly greater than a */
int MixedIntegerES::transformZ(int x, int a, int b)
{
  int x1 = (int) floor(transformR(x, a, b));

#ifdef DEBUG
  cout << "MixedIntegerES::transformZ(): x=" << x << ", a=" << a << ", b=" << b << ", x1=" << x1 << endl; 
#endif

  return x1;
}

/* Fixed convergence for supplied stepsize or probability */ // TODO: change to determinstic schedule from Bck, 1996 - Intelligent Mutation
double MixedIntegerES::fixedConvergence(double x, double threshold, double varPart)
{
  double a = 3.0; // set a to generations*100 for normal reflection (i.e, for all generations, maxValue = varPart + threshold)
  double maxValue = (1 - pow((double)currGen/(generations+1), a)) * varPart + threshold; // generations+1 to make sure maxValue is always > threshold
                                                                                        
#ifdef DEBUG
  cout << "MixedIntegerES::fixedConvergence(): " << maxValue << endl; 
#endif

  if (x < threshold)
    x = threshold;
  else if (x > maxValue)
    x = maxValue;	  

  return x;
}	

//------------------------------------------------------------------------------
// Class Individual
//-------------------------------------------------
// Public Members

Individual::Individual(unsigned n_r, unsigned n_z, unsigned n_d, unsigned n_sigma_r, unsigned n_sigma_z, unsigned n_prob, unsigned n_f, 
		        vector<double>& lBound_r, vector<double>& uBound_r, vector<int>& lBound_z, vector<int>& uBound_z, 
		          vector<int>& lBound_d, vector<int>& uBound_d, double prob_min, double prob_max,
			    double initialSigma_r, double initialSigma_z, double initialProb)
{
  /* Initialize object variables */
  for (unsigned i = 0; i < n_r; i++) // Real-valued: random double within [lBound_r[i],uBound_r[i]]
    R.push_back(uniReal() * (uBound_r[i] - lBound_r[i]) + lBound_r[i]);

  for (unsigned i = 0; i < n_z; i++) // Integer: random int within [lBound_z[i],uBound_z[i]]
    Z.push_back(uniInt() % (uBound_z[i] - lBound_z[i] + 1) + lBound_z[i]);

  for (unsigned i = 0; i < n_d; i++) // Discrete: random int within [lBound_d[i],uBound_d[i]]
    D.push_back(uniInt() % (uBound_d[i] - lBound_d[i] + 1) + lBound_d[i]);

  /* Strategy parameters */
  for (unsigned i = 0; i < n_sigma_r; i++) // Stepsizes for real-valued variables: delta/sqrt(n_r) 
    S_r.push_back((initialSigma_r == 0) ? (uBound_r[i] - lBound_r[i]) / sqrt(n_r) : initialSigma_r); // delta = uBound_r[i]-lBound_r[i], i.e., 
                                                                                                     // rough distance to optimum
  for (unsigned i = 0; i < n_sigma_z; i++) // Stepsizes for integer variables: delta/sqrt(n_z) 
    S_z.push_back((initialSigma_z == 0) ? (uBound_z[i] - lBound_z[i]) / sqrt(n_z) : initialSigma_z); // delta = uBound_z[i]-lBound_z[i], i.e., 
                                                                                                     // rough distance to optimum
  for (unsigned i = 0; i < n_prob; i++) // Mutation probabilities for discrete variables
    Prob.push_back((initialProb == 0) ? uniReal() * (prob_max - prob_min) + prob_min : initialProb - 1.0/n_d); 

  /* Fitness values */
  F.assign(3*n_f+1, 0); // Reserve space for, in case of
                        // A posteriori method: normalized+penalized F values, (sum/product), normalized F values, original F values 
                        // A priori method:     (norm+pen), sum/product of normalized F values, normalized F values, original F values
		        // (see SingleAisle_MIES)  

  /* Initialize feasibility flag */
  feasible = true;

  /* Initialize age */
  age = 0;

  /* Initialize number of Individuals by which this Individual is dominated */
  numDominating = 0;
}

Individual::Individual(vector<double>& R_, vector<int>& Z_, vector<int>& D_, 
	                vector<double>& S_r_, vector<double>& S_z_, vector<double>& Prob_,
                          vector<double>& F_, bool feasible_)
{
  /* Object variables */
  R = R_; Z = Z_; D = D_;

  /* Strategy parameters */ 
  S_r = S_r_; S_z = S_z_; Prob = Prob_;

  /* Fitness values */
  F = F_;

  /* Feasibility flag */
  feasible = feasible_;

  /* Initialize age */
  age = 0;

  /* Initialize number of Individuals by which this Individual is dominated */
  numDominating = 0;
}

Individual::Individual(Individual* original)
{
  if (original != NULL)
  {
    /* Object variables */
    R = original->R; 
    Z = original->Z; 
    D = original->D;

    /* Strategy parameters */ 
    S_r = original->S_r; 
    S_z = original->S_z; 
    Prob = original->Prob;

    /* Fitness values */
    F = original->F;
  }

  /* Feasibility flag */
  feasible = true;

  /* Initialize age */
  age = 0;

  /* Initialize number of Individuals by which this Individual is dominated */
  numDominating = 0;
}

void Individual::dump()
{
  int size = R.size();
  if (size > 0)
    cout << "Real-valued object variables" << endl;
  for (int i = 0; i < size; i++)
  {
    cout << "r" << i << "=" << R[i] << endl;
  }

  size = Z.size();
  if (size > 0)
    cout << "Integer object variables" << endl;
  for (int i = 0; i < size; i++)
  {
    cout << "z" << i << "=" << Z[i] << endl;
  }

  size = D.size();
  if (size > 0)
    cout << "Discrete object variables" << endl;
  for (int i = 0; i < size; i++)
  {
    cout << "d" << i << "=" << D[i] << endl;
  }

  size = S_r.size();
  if (size > 0)
    cout << "Strategy parameters for real-valued object variables" << endl;
  for (int i = 0; i < size; i++)
  {
    cout << "S_r[" << i << "]=" << S_r[i] << endl;
  }

  size = S_z.size();
  if (size > 0)
    cout << "Strategy parameters for integer object variables" << endl;
  for (int i = 0; i < size; i++)
  {
    cout << "S_z[" << i << "]=" << S_z[i] << endl;
  }

  size = Prob.size();
  if (size > 0)
    cout << "Strategy parameters for discrete object variables" << endl;
  for (int i = 0; i < size; i++)
  {
    cout << "Prob[" << i << "]=" << Prob[i] << endl;
  }

  cout << "Age=" << age << endl; 

  cout << "numDominating=" << numDominating << endl; 
}

