#include "NSGA_II.h"

//------------------------------------------------------------------------------
// Class NSGA_II
//-------------------------------------------------
// Public Members

NSGA_II::NSGA_II(unsigned n_f__) 
{ 
#ifdef DEBUG
  cout << "NSGA_II::NSGA_II()" << endl;
#endif

  selectDimension = 3;
  selectFunction.clear();

  for (unsigned i = 0; i < selectDimension; i++)
    selectFunction.push_back(i);

  n_f_ = n_f__;
  nadir.resize(n_f_);
  ideal.resize(n_f_);

  dpSelection = false;
  fixedNadirIdeal = false;
}

/* Select mu Individuals from O and optionally P to form new population P, 
 * using multiobjective selection scheme as proposed for NSGA-II by Deb, 2000
 */
void NSGA_II::select(vector<Individual*>& P, vector<Individual*>& O, unsigned mu, unsigned kappa, unsigned lambda, vector<int>& direction, 
                      vector<double>& bestF)
{
#ifdef DEBUG
  cout << "NSGA_II::select()" << endl;
#endif

  /* Partition population into fronts */
  vector<Individual*> Q;
  int size = 0;
  for (unsigned i = 0; i < lambda; i++)
  {
    if (O[i] != NULL)	  
    {
      Q.push_back(O[i]);
      O[i] = NULL;
      size++;
    }
  }

  /* Select Individuals from parent population P that do not exceed maximum life span */
  for (unsigned i = 0; i < mu; i++)
  {
    if (P[i]->age <= kappa)
    {   
      Q.push_back(P[i]);
      P[i] = NULL;
      size++;
    } 
  }
  
  /* Cleanup unused Individuals from parent population P */
  for (unsigned i = 0; i < mu; i++)
  {  
    delete P[i];
    P[i] = NULL;
  }

  vector<vector<int> > front;
  fastNondominatedSort(Q, front, direction, bestF); 

  /* Select Individuals to form new population: */
  /* 1) try using dominating points selection */
  bool ready = false;
  if (dpSelection) 
  {
    vector<int> indices; // Indices of Individuals in Q
    for (int i = 0; i < size; i++)
      indices.push_back(i);	  

    /* Sort indices based on number of dominating Individuals, in descending order */
    quickSort(indices, 0, size-1, &Q, -1, NULL);

    /* Determine whether enough Individuals can be discarded using dp selection: */
    /* count number of dominated Individuals */
    int numDomIndividuals = 0;
    for (int i = 0; i < size; i++)
    {
      /* Stop counting at first non-dominated Individual */
      if (Q[indices[i]]->numDominating == 0)
        break;
      
      numDomIndividuals++;
    }

    int numToDiscard = size - mu;
    if (numDomIndividuals >= numToDiscard)  
    {
#ifdef DEBUG
  cout << "NSGA_II::select(), using dominating points selection" << endl;
#endif

      /* Copy best Individuals to new population (best are at the back of indices) */
      P.clear();	    
      for (unsigned i = 0; i < mu; i++)
      {
        P.push_back(Q[indices[size-1 - i]]);         
        Q[indices[size-1 - i]] = NULL;
      }

      ready = true; 
    }	
  }
  
  /* 2) or use selection based on fronts */
  if (!ready)
  {
#ifdef DEBUG
  cout << "NSGA_II::select(), using front based selection" << endl;
#endif

    int PSize, frontSize, remaining, numFronts = front.size();
    P.clear();
    for (int i = 0; i < numFronts; i++)
    {
      PSize = P.size();
      frontSize = front[i].size();

      if ((unsigned) PSize == mu)
        break;

      if ((unsigned) (PSize + frontSize) <= mu) // Entire front still fits in population
      {	    
        remaining = frontSize;
      }
      else                         // Use a metric to discriminate between Individuals 
      {	                           // in same front, sort in descending order
        remaining = mu - PSize;

        if (selectDimension > 1)
          frontSort(front[i], Q, direction);
      }

      /* Add Individuals from front[i] to population */
      for (int j = 0; j < remaining; j++)
      {
        P.push_back(Q[front[i][j]]);
        Q[front[i][j]] = NULL;
      }
    }
  }

  /* Cleanup unused Individuals */
  for (int i = 0; i < size; i++)
    delete Q[i];
}

//------------------------------------------------------------------------------
// Class NSGA_II
//-------------------------------------------------
// Protected Members

/* Divide Individuals from O and optionally P in fronts based on domination */
void NSGA_II::fastNondominatedSort(vector<Individual*>& Q, vector<vector<int> >& front, vector<int>& direction, vector<double>& bestF)
{
#ifdef DEBUG
  cout << "NSGA_II::fastNondominatedSort()" << endl;
#endif

  int size = Q.size();
  unsigned tempNumDom;
  map<int, vector<int> > dominates; 
  map<int, unsigned> numDominating;
  vector<int> tempDom, tempFront;
  MOComparison result;

  for (int i = 0; i < size; i++)
  {
    tempDom.clear();
    tempNumDom = 0;

    for (int j = 0; j < size; j++)
    {
      if (i == j)
	continue;

      result = compare(Q[i], Q[j], direction, bestF, false, false); 

      if (result == _Dominating)
	tempDom.push_back(j); // Add j to Individuals which i dominates
      else if (result == _Dominated)
        tempNumDom++; // Increment number of Individuals by which i is dominated
    }

    /* No Individual dominates i, so add i to first front */
    if (tempNumDom == 0)
      tempFront.push_back(i);	    

    dominates[i] = tempDom; // Individuals which i dominates
    numDominating[i] = Q[i]->numDominating = tempNumDom; // Number of Individuals by which i is dominated
  }

  front.push_back(tempFront); // Insert first front into vector front

  int k = 0, frontSize = front[0].size();
  while (!frontSize == 0)
  {
    /* Initialize new front */
    tempFront.clear();

    /* For each Individual j dominated by Individual i in the current front k: */
    for (int i = 0; i < frontSize; i++)
    {
      size = dominates[front[k][i]].size();	    
      for (int j = 0; j < size; j++)
      {
	numDominating[dominates[front[k][i]][j]]--; // Decrease number of Individuals by which j is dominated

        /* If j no longer dominated when not considering the current front: */ 
	if (numDominating[dominates[front[k][i]][j]] == 0)
	  tempFront.push_back(dominates[front[k][i]][j]); // Add j to the new front
      }
    }

    k++;
    front.push_back(tempFront);
    frontSize = front[k].size();
  }

  front.pop_back(); // Remove empty last front
}

/* Determine domination relationship between A and B, or 
 * determine best or worst values */
MOComparison NSGA_II::compare(Individual* A, Individual* B, vector<int>& direction, vector<double>& bestWorstF, bool determineValues, bool worst)
{
  bool better = false,
       worse  = false;

  /* Compare all fitness values (i.e., objective values) */
  for (unsigned i = 0; i < selectDimension; i++)
  {
    if (direction[selectFunction[i]] == 1 || (direction[selectFunction[i]] == -1 && worst)) // Objective to be maximized
    {
      if (A->F[selectFunction[i]] > B->F[selectFunction[i]])
      {
        better = true;

	if (A->F[selectFunction[i]] > bestWorstF[selectFunction[i]] && (A->feasible || determineValues))
	  bestWorstF[selectFunction[i]] = A->F[selectFunction[i]];	
      }
      else if (A->F[selectFunction[i]] < B->F[selectFunction[i]])
      {
	worse = true;

	if (B->F[selectFunction[i]] > bestWorstF[selectFunction[i]] && (B->feasible || determineValues))
	  bestWorstF[selectFunction[i]] = B->F[selectFunction[i]];	
      }
      else
      {
	if (A->F[selectFunction[i]] > bestWorstF[selectFunction[i]] && (A->feasible || determineValues))
	  bestWorstF[selectFunction[i]] = A->F[selectFunction[i]];	
      }
    }
    else if (direction[selectFunction[i]] == -1 || (direction[selectFunction[i]] == 1 && worst)) // Objective to be minimized
    {
      if (A->F[selectFunction[i]] < B->F[selectFunction[i]])
      {
        better = true;

	if (A->F[selectFunction[i]] < bestWorstF[selectFunction[i]] && (A->feasible || determineValues))
	  bestWorstF[selectFunction[i]] = A->F[selectFunction[i]];	
      }
      else if (A->F[selectFunction[i]] > B->F[selectFunction[i]])
      {
	worse = true;

	if (B->F[selectFunction[i]] < bestWorstF[selectFunction[i]] && (B->feasible || determineValues))
	  bestWorstF[selectFunction[i]] = B->F[selectFunction[i]];	
      }
      else
      {
	if (A->F[selectFunction[i]] < bestWorstF[selectFunction[i]] && (A->feasible || determineValues))
	  bestWorstF[selectFunction[i]] = A->F[selectFunction[i]];	
      }
    }

    if (better && worse && !determineValues)
      break;
  }

  if (worse) 
  {
    if (better)
      return _Nondominated;
    else
      return _Dominated;
  }
  else 
  {
    if (better)
      return _Dominating;
    else
      return _Equals;
  }
}

/* Sort Individuals within same front: nD crowding distance sorting (see NSGA-II by Deb, 2000) */
void NSGA_II::frontSort(vector<int>& currFront, vector<Individual*>& Q, vector<int>& direction)
{
#ifdef DEBUG
  cout << "NSGA_II::frontSort(), nD crowding distance sorting" << endl;
#endif

  int frontSize = currFront.size();
  map<int, double> distance;
  double scale;

  for (int i = 0; i < frontSize; i++)
    distance[currFront[i]] = 0;

  /* Sort for each fitness value (i.e., objective value) */
  for (unsigned i = 0; i < selectDimension; i++)
  {
    /* Sort front on objective i, in descending order */
    quickSort(currFront, 0, frontSize-1, &Q, selectFunction[i], NULL); 

    /* Boundary points are always kept */
    distance[currFront[0]] = DBL_MAX;
    distance[currFront[frontSize-1]] = DBL_MAX;

    /* Determine scale for normalizing crowding distance values */
    if (fixedNadirIdeal)
    { 
      // f_max - f_min	    
      if (direction[selectFunction[i]] == 1)
        scale = ideal[selectFunction[i]] - nadir[selectFunction[i]];  
      else	     
        scale = nadir[selectFunction[i]] - ideal[selectFunction[i]];    
    }
    else
      scale = Q[currFront[0]]->F[selectFunction[i]] - Q[currFront[frontSize-1]]->F[selectFunction[i]]; // f_max - f_min, of currFront  
    
    /* For rest of the points, determine crowding distance contribution based on objective i */
    for (int j = 1; j < frontSize-1; j++) 
    {
      if (scale != 0)	    
        distance[currFront[j]] += (Q[currFront[j-1]]->F[selectFunction[i]] - Q[currFront[j+1]]->F[selectFunction[i]]) / scale;  // Large neighbour minus small neighbour
    }
  }

  /* Sort front index currFront on crowding distance, in descending order */
  quickSort(currFront, 0, frontSize-1, NULL, -1, &distance);
}

/* Sort Individuals in front based on objective or number of dominating Individuals or crowding distance, in descending order */
void NSGA_II::quickSort(vector<int>& currFront, int top, int bottom, 
		                  vector<Individual*>* Q, int objective, map<int, double>* distance)
{
#ifdef DEBUG
  cout << "NSGA_II::quickSort(), sorting on " << ((Q != NULL) ? ((objective >= 0) ? "objective " + itos(objective) : "numDominating") 
		                                    : "distance") << endl;
#endif

  /* top = subscript of beginning of vector being considered
   * bottom = subscript of end of vector being considered
   */
  int middle;
  if (top < bottom)
  {
    middle = partition(currFront, top, bottom, Q, objective, distance);
    quickSort(currFront, top, middle, Q, objective, distance); // Sort top partition
    quickSort(currFront, middle+1, bottom, Q, objective, distance); // Sort bottom partition
  }
}

int NSGA_II::partition(vector<int>& currFront, int top, int bottom, 
		                 vector<Individual*>* Q, int objective, map<int, double>* distance)
{
#ifdef DEBUG
  cout << "NSGA_II::partition(), top=" << top << ", bottom=" << bottom << endl;
#endif

  int temp, 
      i = top - 1, 
      j = bottom + 1;

  double x = (Q != NULL) ? ((objective >= 0) ? Q->at(currFront[top])->F[objective] : Q->at(currFront[top])->numDominating) 
	       : distance->find(currFront[top])->second;

  do
  {
    do     
    {
      j--;
    } while (x > ((Q != NULL) ? ((objective >= 0) ? Q->at(currFront[j])->F[objective] : Q->at(currFront[j])->numDominating) 
	            : distance->find(currFront[j])->second));

    do  	
    {
      i++;
    } while (x < ((Q != NULL) ? ((objective >= 0) ? Q->at(currFront[i])->F[objective] : Q->at(currFront[i])->numDominating) 
		    : distance->find(currFront[i])->second));

    if (i < j)
    {     
      /* Switch elements at positions i and j */
      temp = currFront[i];    
      currFront[i] = currFront[j];
      currFront[j] = temp;
    }
  } while (i < j);    
  
  /* Return middle index */
  return j;           
}

/* Determine Nadir and Ideal point for Individuals in currFront */
void NSGA_II::determineNadirIdeal(vector<int>& currFront, int frontSize, vector<Individual*>& Q, vector<int>& direction)
{
#ifdef DEBUG
  cout << "NSGA_II::determineNadirIdeal()" << endl;
#endif

  if (fixedNadirIdeal)
    return;

  /* Reset Ideal and Nadir point */
  for (unsigned i = 0; i < selectDimension; i++)
  {	    
    if (direction[selectFunction[i]] == 1)
    {
      ideal[selectFunction[i]] = 0;
      nadir[selectFunction[i]] = DBL_MAX;
    }
    else
    {
      ideal[selectFunction[i]] = DBL_MAX;
      nadir[selectFunction[i]] = 0;
    }
  } 

  /* Determine Ideal and Nadir point by comparing all Individuals in currFront and respectively selecting best or worst value occurring per objective */
  for (int i = 0; i < frontSize; i++)
  {
    for (int j = 0; j < frontSize; j++) 
    {
      if (i == j)
        continue;

      compare(Q[currFront[i]], Q[currFront[j]], direction, ideal, true, false);      
      compare(Q[currFront[i]], Q[currFront[j]], direction, nadir, true, true);      
    }
  }
}

