#ifndef NSGA_
#define NSGA_

#include <map>
#include "MixedIntegerES.h"

/* Possible relations between Individuals in multiple objective space */
typedef enum {_Dominating, _Dominated, _Nondominated, _Equals} MOComparison; 

class NSGA_II
{
  public:
    NSGA_II(unsigned n_f__ = 3);
    virtual ~NSGA_II() {}

    /** \brief  Input fixed boundary point and fixed ideal point, to be used instead of nadir and ideal point of currFront respectively */
    void setNadirIdeal(vector<double>& nadir_, vector<double>& ideal_) 
    {     
      fixedNadirIdeal = true; 
      nadir = nadir_; 
      ideal = ideal_;
    } 

    /** \brief  Select mu Individuals from O and optionally P to form new population P, using multiobjective selection scheme as proposed for NSGA-II by Deb, 2000 */
    void select(vector<Individual*>& P, vector<Individual*>& O, unsigned mu, unsigned kappa, unsigned lambda,
		 vector<int>& direction, vector<double>& bestF);     
  protected:
    /** \brief  Number of fitness functions used for selection */
    unsigned n_f_, selectDimension;

    /** \brief  Ids of the fitness functions used for selection */  
    vector<unsigned> selectFunction;

    vector<double> nadir, ideal;

    /** \brief  Use dominating points selection, see Beume, 2007 - SMS-EMOA */ 
    bool fixedNadirIdeal, dpSelection;  
                                                                              
    /** \brief  Divide Individuals from O and optionally P in fronts based on domination */
    void fastNondominatedSort(vector<Individual*>& Q, vector<vector<int> >& front, vector<int>& direction, vector<double>& bestF); 

    /** \brief  Determine domination relationship between A and B, or determine best or worst values */
    MOComparison compare(Individual* A, Individual* B, vector<int>& direction, vector<double>& bestWorstF, bool determineValues, bool worst); 
    
    /** \brief  Sort Individuals within same front: nD crowding distance sorting (see NSGA-II by Deb, 2000) */
    virtual void frontSort(vector<int>& currFront, vector<Individual*>& Q, vector<int>& direction); 

    /** \brief  Determine Nadir and Ideal point for Individuals in currFront */
    void determineNadirIdeal(vector<int>& currFront, int frontSize, vector<Individual*>& Q, vector<int>& direction);

    /** \brief  Sort Individuals in front based on objective or number of dominating Individuals or crowding distance, in descending order */
    void quickSort(vector<int>& currFront, int top, int bottom,                         
		    vector<Individual*>* Q, int objective, map<int, double>* distance); 
    int partition(vector<int>& currFront, int top, int bottom, 
		   vector<Individual*>* Q, int objective, map<int, double>* distance);
};

#endif
