#include "TestFunctions_MIES.h"

//------------------------------------------------------------------------------
// Class TestFunctions_MIES
//-------------------------------------------------
// Protected Members

/* Run simulator using the object variables contained in Individual I */
void TestFunctions_MIES::simulate(Individual* I)
{
#ifdef DEBUG
  cout << "TestFunctions_MIES::simulate()" << endl;
#endif

  if (MOproblem != " ")
  {  
#ifdef MSCTHES	  
    if (MOproblem == "DTLZ1")
      DTLZ1(I); 
    else if (MOproblem == "DTLZ2")
      DTLZ2(I);

    for (int i = 0; i < 3; i++) 
      I->F[i] = F[i];
#endif
  }
  else
  {
    I->F[0] = f1(I);
    I->F[1] = f2(I);
    I->F[2] = f3(I);
    I->F[3] = f_sphere(I, n_r, n_z, n_d);
    I->F[4] = f_barrier(I, 100);
  }

  /* Update sum of F values */
  I->F[n_f] = 0;
  for (unsigned i = 0; i < selectDimension; i++)
    I->F[n_f] += I->F[selectFunction[i]];

  if (feedback)
  {
    cout << endl << "Results: " << endl;

    for (unsigned i = 0; i < n_f; i++)
      cout << "F[" << i << "]=" << I->F[i] << endl;
  }
}

/* f1: modified step function             (see Mixed-Integer Evolution Strategy by Emmerich, 2000) */
double TestFunctions_MIES::f1(Individual* I)
{
#ifdef DEBUG
  cout << "TestFunctions_MIES::f1()" << endl;
#endif

  double result = 0;

  for (unsigned i = 0; i < n_r; i++)
    result += pow(floor(I->R[i]), 2);

  for (unsigned i = 0; i < n_z; i++)
    result += pow((double) ((int) (I->Z[i] / 10)), 2);

  for (unsigned i = 0; i < n_d; i++)
    result += pow((double) (I->D[i] % 2), 2);

  return result;
}

/* f2: weighted sphere function           (see Mixed-Integer Evolution Strategy by Emmerich, 2000) */
double TestFunctions_MIES::f2(Individual* I)
{
#ifdef DEBUG
  cout << "TestFunctions_MIES::f2()" << endl;
#endif

  double result = 0;

  for (unsigned i = 0; i < n_r; i++)
    result += (i+1) * pow(I->R[i], 2);

  for (unsigned i = 0; i < n_z; i++)
    result += (i+1) * pow((double) I->Z[i], 2);

  for (unsigned i = 0; i < n_d; i++)
    result += (i+1) * pow((double) I->D[i], 2);

  return result;
}

/* f3: general quadratic fitness function (see Mixed-Integer Evolution Strategy by Emmerich, 2000) */
double TestFunctions_MIES::f3(Individual* I)
{
#ifdef DEBUG
  cout << "TestFunctions_MIES::f3()" << endl;
#endif

  double result = 0, temp = 0;
 
  for (unsigned i = 0; i < n_r; i++)
  {
    temp = 0;
    for (unsigned j = 0; j <= i; j++)
      temp += I->R[j] + I->Z[j] + I->D[j];

    result += pow(temp, 2);
  }

  return result;
}

/* f_sphere: generalized sphere function  (see Mixed-Integer Evolution Strategies by Li, 2006) */
double TestFunctions_MIES::f_sphere(Individual* I, unsigned n_r, unsigned n_z, unsigned n_d)
{
#ifdef DEBUG
  cout << "TestFunctions_MIES::f_sphere()" << endl;
#endif

  double result = 0;

  for (unsigned i = 0; i < n_r; i++)
    result += pow(I->R[i], 2);

  for (unsigned i = 0; i < n_z; i++)
    result += pow((double) I->Z[i], 2);

  for (unsigned i = 0; i < n_d; i++)
    result += pow((double) I->D[i], 2);

  return result;
}

/* f_barrier: multimodal barrier problem  (see Mixed-Integer Evolution Strategies by Li, 2006) */
double TestFunctions_MIES::f_barrier(Individual* I, int C)
{
#ifdef DEBUG
  cout << "TestFunctions_MIES::f_barrier()" << endl;
#endif

  double result = 0;
  static vector<int> A;
  int j, temp;

  if (A.empty())
  { 
    for (int i = 0; i < 21; i++)
      A.push_back(i);

    for (int k = 0; k < C; k++)
    {
      j = uniInt() % 20;
      temp = A[j];
      A[j] = A[j+1];
      A[j+1] = temp;
    }
  }

  result += f_sphere(I, n_r, 0, n_d);

  for (unsigned i = 0; i < n_z; i++)
    result += pow((double) A[I->Z[i]], 2);

  return result;
}

/* Write current population P to file */
void TestFunctions_MIES::writePop(time_t id)
{
#ifdef DEBUG
  cout << "TestFunctions_MIES::writePop()" << endl;
#endif

  /* Write current population */	
  string filename = "RunID_" + itos((int) id) + "_P.txt"; 
  fstream outfile(filename.c_str(), ios::out);
 	
  int size;
  for (int k = 0; k < 2; k++)
  {
    for (unsigned i = 0; i < mu; i++)
    {
      outfile << ((P[i]->feasible) ? "F" : "P" ) << "\t";

      size = P[i]->F.size();	  
      for (int j = 0; j < size; j++)
        outfile << P[i]->F[j] << "\t";

      outfile << "\t\t\t\t";
            
      for (unsigned j = 0; j < n_r; j++)
        outfile << P[i]->R[j] << "\t";

      for (unsigned j = 0; j < n_z; j++)
        outfile << P[i]->Z[j] << "\t";

      for (unsigned j = 0; j < n_d; j++)
        outfile << P[i]->D[j] << "\t";

      outfile << "\t\t\t\t";

      for (unsigned j = 0; j < n_sigma_r; j++)
        outfile << P[i]->S_r[j] << "\t";

      for (unsigned j = 0; j < n_sigma_z; j++)
        outfile << P[i]->S_z[j] << "\t";

      for (unsigned j = 0; j < n_prob; j++)
        outfile << P[i]->Prob[j] << "\t";

      outfile << endl << endl << endl; // Two white lines to indicate new gnuplot index
    }
    outfile.close();

    if (k < 1)
    {
      /* Only for first generation write population to permanent file */
      if (currGen > 0) // Comment out these lines
	break;         // for all generations to be written to files

      filename = "RunID_" + itos((int) id) + "_P_g" + itos(currGen) + ".txt"; 
      outfile.open(filename.c_str(), ios::out);
    }
  }

  /* Output stepsizes */
  filename = "RunID_" + itos((int) id) + "_stepsizes.txt"; 
  outfile.open(filename.c_str(), ios::out | ios::ate | ios::app);
  outfile << currGen << "\t"; 

  for (unsigned i = 0; i < mu; i++)
  {
    for (unsigned j = 0; j < n_sigma_r; j++)
      outfile << P[i]->S_r[j] << "\t";

    for (unsigned j = 0; j < n_sigma_z; j++)
      outfile << P[i]->S_z[j] << "\t";

    for (unsigned j = 0; j < n_prob; j++)
      outfile << P[i]->Prob[j] << "\t";

    outfile << "\t\t\t\t";
  }

  outfile << endl;
  outfile.close();

  /* Determine bestF for initial population */
  if (currGen == 0) 
    determineBestF(); 

  /* Append best fitness values to file */
  filename = "RunID_" + itos((int) id) + "_bestF.txt"; 
  outfile.open(filename.c_str(), ios::out | ios::ate | ios::app);

  outfile << currGen << "\t";

#ifdef APRIORI
  outfile << bestF[selectFunctionActual];
#else
  for (unsigned i = 0; i < selectDimension; i++) 
    outfile << bestF[selectFunction[i]] << "\t";
#endif

  outfile << endl;

  outfile.close();
}

/* Log time that was needed to complete generation */
void TestFunctions_MIES::writeLog(time_t elapsed, time_t id)
{
#ifdef DEBUG
  cout << "TestFunctions_MIES::writeLog()" << endl;
#endif

  string filename = "RunID_" + itos((int) id) + "_log.txt"; 
  fstream outfile(filename.c_str(), ios::out | ios::ate | ios::app);

  if (currGen <= generations)
  {
    if (feedback)
      cout << endl << "Generation " << currGen << " took " << elapsed << "s" << endl;
    outfile << "Generation " << currGen << ": \t" << elapsed << "s" << endl;
  }
  else
  {
    if (feedback)
      cout << "Optimization took " << elapsed << "s" << endl << endl;
    outfile << endl << "Total optimization: \t" << elapsed << "s" << endl;
  }

  outfile.close();
}

